"""Logger config."""

import contextlib
import logging
import os
import stat
import sys

from loguru import logger


class InterceptHandler(logging.Handler):
    """See https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging."""  # noqa

    def emit(self, record: logging.LogRecord) -> None:
        """Redirect standard logging to loguru sink."""
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = str(record.levelno)

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            assert frame.f_back
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


logging.basicConfig(handlers=[InterceptHandler()], level=0)


def configure_logger() -> None:
    """Logger config."""

    logger.remove()
    logger.add(sys.stderr, level="TRACE", colorize=False)

    with contextlib.suppress(OSError):
        if stat.S_ISREG(os.stat(3).st_mode):
            logger.add(os.fdopen(3, mode="w"), level="TRACE", colorize=True)

    add_logging_levels()


def add_logging_levels() -> None:
    """Add custom logging levels."""

    # pylint: disable=too-many-statements

    # remove bold from loguru default colors
    for lvl in logger._core.levels.values():  # type: ignore # noqa: protected-access
        logger.level(lvl.name, color=lvl.color.replace("<bold>", ""))

    # set severity of custom levels relative to the builtins
    _always = _warn = logger.level("WARNING").no
    _info = logger.level("INFO").no
    _debug = logger.level("DEBUG").no
    _trace = logger.level("TRACE").no

    # black blue cyan green magenta red white yellow
    # bold dim normal hide italic blink strike underline reverse
    # logger.level('DEBUG', no=_debug, color='<cyan>')
    # logger.level('TRACE', color='<white>')

    # admin
    logger.level("ADMIN", no=_always, color="<magenta><bold><italic>")
    logger.level("KICK", no=_always, color="<red><italic><reverse>")

    # users
    logger.level("ADDUSER", no=_always, color="<yellow><italic>")
    logger.level("INACTIVE", no=_always, color="<white>")

    # lobby
    logger.level("ADDLOBBY", no=_trace, color="<yellow><italic><bold>")

    # queues
    logger.level("CLEAR", no=_info, color="<magenta>")
    logger.level("EMPTY", no=_info, color="<magenta><reverse>")
    logger.level("PUSH", no=_info, color="<magenta><italic>")
    logger.level("PUSHLEFT", no=_info, color="<magenta><italic><bold>")
    logger.level("POP", no=_info, color="<magenta><italic>")
    logger.level("POPLEFT", no=_info, color="<magenta><italic><bold>")

    logger.level("SPAMS", no=_always, color="<white>")
    logger.level("KICKS", no=_always, color="<white><italic>")

    # gameplay
    logger.level("CAPBLU", no=_always, color="<cyan>")
    logger.level("CAPRED", no=_always, color="<red>")
    logger.level("CHAT", no=_always, color="<green>")
    logger.level("CHATBLU", no=_always, color="<cyan>")
    logger.level("CHATRED", no=_always, color="<red>")
    logger.level("CONNECT", no=_always, color="<magenta>")
    logger.level("DEFBLU", no=_always, color="<cyan>")
    logger.level("DEFRED", no=_always, color="<red>")
    logger.level("KILL", no=_always, color="<green>")
    logger.level("KILLBLU", no=_always, color="<cyan>")
    logger.level("KILLRED", no=_always, color="<red>")
    logger.level("STATUS", no=_trace, color="<magenta>")
    logger.level("SUICIDE", no=_always, color="<red>")
    logger.level("TEAMCHAT", no=_always, color="<green><italic>")
    logger.level("TEAMCHATBLU", no=_always, color="<cyan><italic>")
    logger.level("TEAMCHATRED", no=_always, color="<red><italic>")
    logger.level("PERK-ON", no=_always, color="<green><italic>")
    logger.level("PERK-OFF", no=_always, color="<green>")

    # logger.level('action', no=_debug,  color='<blue><bold>')
    # logger.level('assign', no=_debug,  color='<green>')
    logger.level("div", no=_debug, color="<cyan>")
    logger.level("duel", no=_info, color="<magenta>")
    logger.level("exclude", no=_trace, color="<white>")
    logger.level("help", no=_always, color="<magenta>")
    logger.level("ignore", no=_debug, color="<white>")
    logger.level("injected", no=_always, color="<yellow><reverse><italic>")
    logger.level("logline", no=_debug, color="<yellow>")
    logger.level("nextline", no=_warn, color="<yellow>")
    logger.level("console", no=_always, color="<green><bold>")
    logger.level("Player", no=_always, color="<yellow><reverse>")
    logger.level("SteamPlayer", no=_debug, color="<green>")
    logger.level("regex", no=_trace, color="<magenta>")
    logger.level("report", no=_always, color="<cyan>")
    logger.level("server", no=_trace, color="<cyan>")
    logger.level("QVALVE", no=_always, color="<cyan><italic>")
    logger.level("user", no=_debug, color="<green>")
    logger.level("RED", no=_debug, color="<red>")
    logger.level("BLU", no=_debug, color="<cyan>")

    logger.level("BOT", no=_always, color="<cyan><reverse>")
    logger.level("FRIENDS", no=_always, color="<cyan><reverse>")
    logger.level("TACOBOT", no=_always, color="<cyan><reverse>")
    logger.level("PAZER", no=_always, color="<cyan><reverse>")
    logger.level("MILENKO", no=_always, color="<yellow><bold><italic><reverse>")
    logger.level("CHEATER", no=_always, color="<yellow><bold><italic>")
    logger.level("SUSPECT", no=_always, color="<magenta><bold><italic>")
    logger.level("RACIST", no=_always, color="<red><bold><italic><reverse>")
    logger.level("FUZZ", no=_always, color="<yellow><reverse>")
    logger.level("ALIAS", no=_always, color="<cyan><reverse>")
    logger.level("NOTIFY", no=_always, color="<yellow><reverse>")



# import pytest

from tf2mon.texttable import TextColumn, TextTable


def test_column() -> None:

    table = TextTable(
        [
            TextColumn(-10, "STEAMID"),
            TextColumn(4.1, "KD"),
            TextColumn(-4, "AGE"),
            TextColumn(1, "P"),
            TextColumn(2, "CC"),
            TextColumn(2, "SC"),
            TextColumn(4, "CI"),
            TextColumn(25, "PERSONANAME"),
            TextColumn(0, "REALNAME"),
        ]
    )

    assert (
        table.formatted_header
        == "   STEAMID   KD  AGE P CC SC CI   PERSONANAME               REALNAME"
    )

    assert (
        table.format_detail(
            12345,
            3.14159,
            5000,
            "1",
            "2",
            "3",
            "4",
            "Persona Name",
            "Real Name",
        )
        == "     12345  3.1 5000 1 2  3  4    Persona Name              Real Name"
    )



"""Single-step controls."""

import re
import threading
from typing import Callable, Match, Pattern

from loguru import logger

import tf2mon
from tf2mon.control import Control


class SingleStepControl(Control):
    """Single-step control."""

    is_stepping: bool = False
    pattern: Pattern[str] | None = None
    clear: Callable[[], None] | None = None
    set: Callable[[], None] | None = None
    wait: Callable[[float | None], bool] | None = None
    _event: threading.Event | None = None

    def start(self) -> None:
        self._event = threading.Event()
        self.clear = self._event.clear
        self.set = self._event.set
        self.wait = self._event.wait

        if tf2mon.options.breakpoint is not None:
            self.set_single_step_lineno(tf2mon.options.breakpoint)

        if pattern := tf2mon.options.search:
            if pattern.startswith("/"):
                pattern = pattern[1:]
            self.set_single_step_pattern(pattern)

        if tf2mon.options.single_step:
            self.start_single_stepping()
        else:
            self.stop_single_stepping()

    def start_single_stepping(self) -> None:
        self.is_stepping = True
        assert self._event
        self._event.clear()

    def stop_single_stepping(self) -> None:
        self.is_stepping = False
        assert self._event
        self._event.set()

    def set_single_step_lineno(self, lineno: int = 0) -> None:
        """Begin single-stepping at `lineno` if given else at eof."""

        if lineno:
            assert tf2mon.conlog
            tf2mon.conlog.inject_cmd(lineno, "SINGLE-STEP")
        else:
            # stop single-stepping until eof, then single-step again
            self.stop_single_stepping()

    def set_single_step_pattern(self, pattern: str | None = None) -> None:
        """Begin single-stepping at next line that matches `pattern`.

        Pass `None` to clear the pattern.
        """

        if not pattern:
            self.pattern = None
            logger.log("ADMIN", "clear pattern")
        else:
            flags = 0
            if pattern.endswith("/i"):
                pattern = pattern[:-2]
                flags = re.IGNORECASE
            elif pattern.endswith("/"):
                pattern = pattern[:-1]
            self.pattern = re.compile(pattern, flags)
            logger.log("ADMIN", f"set search={self.pattern}")


class SingleStepStartControl(Control):
    """Start single-stepping."""

    name = "SINGLE-STEP"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.SingleStepControl.start_single_stepping()
        logger.info("single-step")



"""Kicks queue control."""

from typing import Match

import tf2mon
from tf2mon.controls.msgqueue import MsgQueueControl


class KicksControl(MsgQueueControl):
    """Kicks queue control."""

    name = "kicks"


class KicksPopControl(KicksControl):
    """Pop last/latest kicks queue message."""

    name = "KICKS-POP"
    action = "tf2mon_kicks_pop"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.KicksControl.pop()
        tf2mon.ui.refresh_kicks()


class KicksClearControl(KicksControl):
    """Clear kicks queue."""

    name = "KICKS-CLEAR"
    action = "tf2mon_kicks_clear"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.KicksControl.clear()
        tf2mon.ui.refresh_kicks()


class KicksPopleftControl(KicksControl):
    """Pop first/oldest kicks queue message."""

    name = "KICKS-POPLEFT"
    action = "tf2mon_kicks_popleft"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.KicksControl.popleft()
        tf2mon.ui.refresh_kicks()



"""Collection of `Control`s."""

import time
from argparse import ArgumentParser

from libcurses import register_fkey
from loguru import logger

# from tf2mon.command import CommandManager
from tf2mon.control import Control
from tf2mon.fkey import FKey
from tf2mon.pkg import APPNAME, APPTAG


class Controller:
    """Collection of `Control`s."""

    # all controls; irrelevant order.
    controls: list[Control] = []

    # subset bound to function-keys; ordered for rendering `--help`.
    bindings: list[Control] = []

    def __init__(self, controls: list[Control] | None = None) -> None:
        """Initialize `Controller` with list of `Control`s."""

        if controls is not None:
            self.controls = controls

    def bind(self, control: Control, keyspec: str) -> None:
        """Bind `control` to `keyspec`."""

        control.fkey = FKey(keyspec)
        control.fkey.bind(control)
        self.bindings.append(control)

    def add_arguments_to(self, parser: ArgumentParser) -> None:
        """Add arguments for all controls to `parser`."""

        Control.cli = parser.get_default("cli")

        for control in [x for x in self.controls if hasattr(x, "add_arguments_to")]:
            control.add_arguments_to(parser)

    def fkey_help(self) -> str:
        """Return help for function keys."""

        lines = []
        for control in self.bindings:
            # 17 == indent 4 + len("KP_RIGHTARROW")
            assert control.fkey
            lines.append(f"{control.fkey.longname:>17} {control.__doc__}")
        return "\n".join(lines)

    def start(self) -> None:
        """Finalize initialization now that curses has been started."""

        for control in self.controls:
            if hasattr(control, "start"):
                logger.debug(f"Starting {control.__class__.__name__}")
                control.start()
            # else:
            #     logger.warning(f"Not starting {control.__class__.__name__}")

        for control in [x for x in self.bindings if x.fkey and not x.fkey.is_ascii]:
            if hasattr(control, "handler"):
                assert control.fkey
                logger.debug(
                    f"Registering {control.__class__.__name__} to {control.fkey.longname}"
                )
                register_fkey(control.handler, control.fkey.key)

    def as_tf2_exec_script(self, static: str, dynamic: str) -> str:
        """Return game key bindings as TF2 exec script."""

        lines = [
            f"// Auto-generated by {APPNAME} on {time.asctime()}",
            "",
            'echo "-------------------------------------------------------------------------"',
            f'echo "Running {static}"',
            "",
            "// Push from game to monitor.",
            'alias tf2mon_push               "status; tf_lobby_debug"',
            "",
            "// Pull from monitor into game.",
            f'alias tf2mon_pull               "exec {dynamic}"',
            "",
            "// which constantly redefines these aliases:",
            'alias tf2mon_kicks_pop          "tf2mon_pull; _tf2mon_kicks_pop"',
            'alias tf2mon_kicks_popleft      "tf2mon_pull; _tf2mon_kicks_popleft"',
            'alias tf2mon_spams_pop          "tf2mon_pull; _tf2mon_spams_pop"',
            'alias tf2mon_spams_popleft      "tf2mon_pull; _tf2mon_spams_popleft"',
            "",
            f'alias tf2mon_kicks_clear        "echo {APPTAG}KICKS-CLEAR"',
            f'alias tf2mon_spams_clear        "echo {APPTAG}SPAMS-CLEAR"',
            f'alias tf2mon_clear_queues       "echo {APPTAG}CLEAR-QUEUES"',
            "",
            "// Bind unmodified keys directly to their `base` commands.",
        ]

        for key in [
            x for x in FKey.pkeys.values() if x.name and x.base and not (x.ctrl or x.shift)
        ]:
            if key.base and hasattr(key.base, "action") and key.base.action:
                lines.append(f'bind "{key.name}" "{key.base.action}"')

        lines.extend(
            [
                "",
                "// Bind modified keys indirectly through aliases.",
            ]
        )

        _base = ["_user_bind_base", "_class_bind_base"]
        _ctrl = ["_user_bind_ctrl", "_class_bind_ctrl"]
        _shift = ["_user_bind_shift", "_class_bind_shift"]

        for key in [x for x in FKey.pkeys.values() if x.name and (x.ctrl or x.shift)]:
            if key.base and hasattr(key.base, "action") and key.base.action:
                alias = f"_b{key.name}"
                lines.append(f'alias {alias} "{key.base.action}"')
                _base.append(f"bind {key.name} {alias}")

            if key.ctrl and hasattr(key.ctrl, "action") and key.ctrl.action:
                alias = f"_c{key.name}"
                lines.append(f'alias {alias} "{key.ctrl.action}"')
                _ctrl.append(f"bind {key.name} {alias}")

            if key.shift and hasattr(key.shift, "action") and key.shift.action:
                alias = f"_s{key.name}"
                lines.append(f'alias {alias} "{key.shift.action}"')
                _shift.append(f"bind {key.name} {alias}")

        lines.extend(
            [
                "",
                "// User class scripts may hook in through these;",
                "// e.g., `.tf2/cfg/user/engineer.cfg` could:",
                '//   alias _class_bind_base "bind 1 slot1; bind 2 slot2" etc.',
                '//   alias _class_bind_shift "bind 1 build_sentry; bind 2 build_dispenser" etc',
                'alias _class_bind_base ""',
                'alias _class_bind_ctrl ""',
                'alias _class_bind_shift ""',
                "",
                "// User scripts may hook in through these.",
                "// e.g., `.tf2/cfg/user/bad-dad-aliases.cfg` could:",
                '//   alias _user_bind_base "bind a +moveleft"',
                '//   alias _user_bind_shift "bind a load_itempreset 0"',
                'alias _user_bind_base ""',
                'alias _user_bind_ctrl ""',
                'alias _user_bind_shift ""',
                "",
                "// Toggle modified keys.",
            ]
        )

        # alias _bind_base "bind f1 _bf1; bind f2 _bf2; bind f3 _bf3; bind f4 _bf4"
        # alias _bind_ctrl "bind f1 _cf1; bind f2 _cf2; bind f3 _cf3; bind f4 _cf4"
        # alias _bind_shift "bind f1 _sf1; bind f2 _sf2; bind f3 _sf3; bind f4 _sf4"

        if binds := "; ".join(_base):
            lines.append(f'alias _bind_base "{binds}"')

        if binds := "; ".join(_ctrl):
            lines.append(f'alias _bind_ctrl "{binds}"')

        if binds := "; ".join(_shift):
            lines.append(f'alias _bind_shift "{binds}"')

        lines.extend(
            [
                "",
                'bind ctrl "+ctrled"',
                "alias +ctrled _bind_ctrl",
                "alias -ctrled _base_base",
                "",
                'bind shift "+shifted"',
                "alias +shifted _bind_shift",
                "alias -shifted _bind_base",
                "",
                "_bind_base",
            ]
        )

        return "\n".join(lines) + "\n"

    def get_status_line(self) -> str:
        """Return 1-line string showing current state of function keys."""

        items: list[str] = []
        for pkey in FKey.pkeys.values():
            for control in pkey.bindings:
                assert control is not None
                if hasattr(control, "status") and control.status:
                    items.append(f"{control.fkey.shortname}={control.status()}")
        return " ".join(items)



"""Table of `Player`s."""

from __future__ import annotations

import json
import time
from dataclasses import dataclass

from loguru import logger

from tf2mon.database import Database, DatabaseTable


@dataclass
class Player(DatabaseTable):
    """A `Player`."""

    # pylint: disable=too-many-instance-attributes

    __tablename__ = "players"

    # database columns
    steamid: int  # primary key

    # defcon6
    bot: str = ""
    friends: str = ""
    tacobot: str = ""
    pazer: str = ""

    # playerlist.official
    _cheater: str = ""
    _suspect: str = ""
    _exploiter: str = ""
    _racist: str = ""
    _last_name: str = ""
    _s_last_time: str = ""

    # tf2mon
    cheater: str = ""
    suspect: str = ""
    exploiter: str = ""
    racist: str = ""
    milenko: str = ""
    last_name: str = ""
    s_last_time: str = ""
    names: str = ""

    # Attributes.
    # base
    CHEATER = "cheater"
    SUSPECT = "suspect"
    EXPLOITER = "exploiter"
    RACIST = "racist"
    # tf2mon
    MILENKO = "milenko"
    GAMEBOT = "gamebot"
    # defcon6
    BOT = "bot"
    FRIENDS = "friends"
    TACOBOT = "tacobot"
    PAZER = "pazer"

    def __post_init__(self) -> None:
        # additional (non-database) properties

        self._s_prev_time = self._s_last_time
        self.s_prev_time = self.s_last_time

    @classmethod
    def create_table(cls) -> None:
        """Execute create table statement."""

        db = Database()
        assert db

        db.execute(
            f"create table if not exists {cls.__tablename__}"
            """(
                steamid integer primary key,
                bot text,
                friends text,
                tacobot text,
                pazer text,
                _cheater text,
                _suspect text,
                _exploiter text,
                _racist text,
                _last_name text,
                _s_last_time text,
                cheater text,
                suspect text,
                exploiter text,
                racist text,
                milenko text,
                last_name text,
                s_last_time text,
                names text
            )""",
        )
        db.connection.commit()

    def getattrs(self) -> list[str]:
        """Docstring."""

        return [
            self.bot,
            self.friends,
            self.tacobot,
            self.pazer,
            self._cheater,
            self._suspect,
            self._exploiter,
            self._racist,
            self._last_name,
            self._s_last_time,
            self.cheater,
            self.suspect,
            self.exploiter,
            self.racist,
            self.milenko,
        ]

    def setattrs(self, attrs: list[str]) -> None:
        """Docstring."""
        for attr in attrs:
            setattr(self, attr, attr)

    @classmethod
    def fetch_steamid(cls, steamid: int) -> Player | None:
        """Return `Player` for given steamid, else None if not found."""

        db = Database()
        assert db

        db.execute(f"select * from {cls.__tablename__} where steamid=?", (steamid,))
        if row := db.fetchone():
            return cls(*tuple(row))
        return None

    def track_appearance(self, name: str) -> None:
        """Record user appearing as given name."""

        level = self.display_level
        leader = f"{level}: {self.steamid}"

        logger.log(level, f"{leader}: name: `{name}`")
        logger.log(level, f"{leader}: last={self.s_last_time} {self._s_last_time}")

        self._s_prev_time = self._s_last_time
        self.s_prev_time = self.s_last_time
        self.s_last_time = self.strftime()

        aliases = list(self.aliases)  # deserialize

        if self.last_name not in aliases:
            logger.log(level, f"{leader}: adding last_name `{self.last_name}`")
            aliases.append(self.last_name)
        self.last_name = name

        if name not in aliases:
            logger.log(level, f"{leader}: adding alias `{name}`")
            aliases.append(name)

        self.names = json.dumps({"json": aliases})  # serialize
        self.upsert()

    @property
    def aliases(self) -> list[str]:
        """Return list of names used by this account."""

        ret: list[str] = json.loads(self.names).get("json", []) if self.names else []
        return ret

    def strftime(self, seconds: int | None = None) -> str:
        """Return time formatted as a string."""

        if not seconds:
            seconds = int(time.time())

        return time.strftime("%FT%T", time.localtime(seconds))

    @property
    def display_level(self) -> str:
        """Return logging level for displaying this player."""

        # pylint: disable=too-many-return-statements

        # Ordered.
        if self.racist or self._racist:
            return "RACIST"
        if self.cheater or self._cheater:
            return "CHEATER"
        if self.suspect or self._suspect:
            return "SUSPECT"
        if self.exploiter or self._exploiter:
            return "EXPLOITER"
        if self.milenko:
            return "MILENKO"
        if self.bot:
            return "BOT"
        if self.friends:
            return "FRIENDS"
        if self.tacobot:
            return "TACOBOT"
        if self.pazer:
            return "PAZER"

        logger.error("unknown attributes")
        return self.__class__.__name__

    @property
    def is_banned(self) -> bool:
        """Return True if hacker is suspect or racist."""

        return bool(
            self.racist
            or self._racist
            or self.cheater
            or self._cheater
            or self.exploiter
            or self._exploiter
            or self.bot
            or self.friends
            or self.tacobot
            # or self.pazer
        )

    @staticmethod
    def new_player(steamid: int, attrs: list[str], name: str) -> Player:
        """Create, insert and return new `Player`."""

        player = Player(steamid)
        player.setattrs(attrs)
        player.track_appearance(name)
        return player



"""Role (job, function, position) of `User`."""

import csv
from enum import Enum
from pathlib import Path
from typing import Tuple

from loguru import logger

Role = Enum("Role", "scout soldier pyro demo heavy engineer medic sniper spy unknown")

_ROLE_BY_WEAPON: dict[str, Role] = {}


def load_weapons_data(path: Path) -> None:
    """Load weapons data from `path`."""

    global _ROLE_BY_WEAPON  # pylint: disable=global-statement

    logger.info(f"Reading `{path}`")
    with open(path, encoding="utf-8") as _f:
        _ROLE_BY_WEAPON = {
            weapon: Role.__dict__[role_name] for role_name, weapon in csv.reader(_f)
        }


def get_role_weapon_state(
    default_role: Role, weapon: str, crit: bool, perk: str
) -> Tuple[Role, str]:
    """Return role and hashable key for given weapon_state.

    Also used as the formatted display value by scoreboard.
    """

    role = _ROLE_BY_WEAPON.get(weapon, default_role)

    s_crit = "+crit" if crit else ""
    s_perk = "+" + perk if perk else ""

    weapon_state = " ".join(
        [
            f"{role.name:8}",  # 8=len("engineer")
            f"{s_crit:5}",  # 5=len("+crit")
            f"{weapon:26}",  # 26=len("tf_projectile_pipe_remote")
            s_perk,  # 20=len("Infinite Double Jump")
        ]
    ).rstrip()

    return role, weapon_state



import pytest

from tf2mon.steamplayer import SteamPlayer

# pylint: disable=unused-argument

VALUEHOLDERS = "values(?,?,?,?,?,?,?,?,?,?)"


def test_valueholders_cls() -> None:
    assert SteamPlayer.valueholders() == VALUEHOLDERS


def test_valueholders_obj() -> None:
    assert SteamPlayer(0).valueholders() == VALUEHOLDERS


def _test_select_all(session: str) -> None:
    for result in SteamPlayer.select_all():
        print(result)


@pytest.mark.parametrize(("steamid"), [-3, -2, -1, 0, 1])
def test_fetch_steamid_not_found(session: str, steamid: int) -> None:
    assert SteamPlayer.fetch_steamid(steamid) is None


@pytest.mark.parametrize(("steamid"), [42708103])
def _test_fetch_steamid_found(session: str, steamid: int) -> None:
    result = SteamPlayer.fetch_steamid(steamid)
    assert result
    print(result)



"""Chat layout."""

import libcurses

from tf2mon.baselayout import BaseLayout


class ChatLayout(BaseLayout):
    """Chat layout.

    +-------------------+----------------------+
    | scorewin_blu      | scorewin_red         |  section 1
    |                   |                      |
    +-------------------+----------------------+
    | chatwin_blu       | chatwin_red          |  section 2
    |                   |                      |
    +--------------+----+----------------------+
    | user_win     | logger_win                |  section 3
    |              |                           |
    +--------------+--------+------------------+
    | status_win            | cmdline_win      |  section 4
    +-----------------------+------------------+
    """

    # pylint: disable=too-many-instance-attributes

    def __post_init__(self, grid: libcurses.Grid, max_users: int) -> None:
        """Build windows."""

        n_score_lines = int(max_users / 2) + 2 + 1  # 2=borders (top and bottom), 1=header.
        # 124=len(self._scoreboard._formatted_header) + 2=borders (left and right) + 1=padding
        # _ncols = min(124, int(2 * grid.ncols / 3))
        _ncols = int(grid.ncols / 2)

        n_avail = (grid.nlines - n_score_lines) - 1  # 1=border (between chat and user).

        min_n_chat_lines = 20
        max_n_user_lines = 20

        if n_avail < min_n_chat_lines + max_n_user_lines:
            n_chat_lines = int(n_avail / 2)
            n_user_lines = n_avail - n_chat_lines
        else:
            n_user_lines = max_n_user_lines
            n_chat_lines = n_avail - n_user_lines

        # A1 top-left
        self.scorewin_blu = grid.box(
            "scorewin_blu",
            top=grid,
            left=grid,
            nlines=n_score_lines,
            ncols=_ncols,
        )

        # B1 top-right
        self.scorewin_red = grid.box(
            "scorewin_red",
            top=grid,
            right=grid,
            nlines=n_score_lines,
            ncols=0,
            left2r=self.scorewin_blu,
        )

        # -------------------------------------
        # A2 leave vertical gap for chatwin_blu
        # -------------------------------------

        # A4 bottom-left
        self.status_win = grid.box(
            "status",
            bottom=grid,
            left=grid,
            nlines=3,
            ncols=grid.ncols * 2 // 3,
        )

        # B4 bottom-right
        self.cmdline_win = grid.box(
            "cmdline",
            bottom=grid,
            right=grid,
            nlines=3,
            ncols=0,
            left2r=self.status_win,
        )

        # A3 above bottom-left
        self.user_win = grid.box(
            "user",
            bottom2t=self.status_win,
            left=grid,
            nlines=n_user_lines,
            ncols=min(62, grid.ncols // 2),  # 62: see get_weapon_state
        )

        # B3 above bottom-right
        self.logger_win = grid.box(
            "logwin",
            bottom=self.user_win,
            right=grid,
            top=self.user_win,
            left2r=self.user_win,
            nlines=0,
            ncols=0,
        )

        # A2 below top-left
        self.chatwin_blu = grid.box(
            "chatwin_blu",
            top2b=self.scorewin_blu,
            left=grid,
            nlines=0,
            ncols=_ncols,
            bottom2t=self.user_win,
        )

        # B2 below top-right
        self.chatwin_red = grid.box(
            "chatwin_red",
            top2b=self.scorewin_red,
            right=grid,
            left2r=self.chatwin_blu,
            bottom2t=self.user_win,
            nlines=0,
            ncols=0,
        )



"""Communication mechanism to "send" data to the game."""

from collections import deque

from loguru import logger

import tf2mon
from tf2mon.pkg import APPTAG


class MsgQueue:
    """Communication mechanism to "send" data to the game."""

    def __init__(self, name: str):
        """Create `MsgQueue` named `name`."""

        self.name = name
        self.msgs: deque[str] = deque()

    def push(self, msg: str) -> None:
        """Append message to end of queue."""

        self.msgs.append(msg)
        logger.opt(depth=1).log("PUSH", msg)

    def pushleft(self, msg: str) -> None:
        """Append message to other end of queue."""

        self.msgs.insert(0, msg)
        logger.opt(depth=1).log("PUSHLEFT", msg)

    def pop(self) -> None:
        """Remove and return message from end of queue."""

        if not self.msgs:
            logger.log("EMPTY", f" {self.name} ".center(80, "-"))
        else:
            logger.opt(depth=1).log("POP", self.msgs.pop())

    def popleft(self) -> None:
        """Remove and return message from other end of queue."""

        if not self.msgs:
            logger.log("EMPTY", f" {self.name} ".center(80, "-"))
        else:
            logger.opt(depth=1).log("POPLEFT", self.msgs.popleft())

    def clear(self) -> None:
        """Remove all messages from the queue."""

        self.msgs.clear()
        logger.opt(depth=1).log("CLEAR", self.name)

    def aliases(self) -> list[str]:
        """Return message queue function key aliases.

        Create and return definitions of the `_tf2mon` prefixed-
        aliases that are called from `~/tf2/cfg/tf2-monitor.cfg`.

                                                      --- these aliases ---
        alias tf2mon_kicks_pop          "tf2mon_pull; _tf2mon_kicks_pop"
        alias tf2mon_kicks_popleft      "tf2mon_pull; _tf2mon_kicks_popleft"
        alias tf2mon_spams_pop          "tf2mon_pull; _tf2mon_spams_pop"
        alias tf2mon_spams_popleft      "tf2mon_pull; _tf2mon_spams_popleft"
        """

        # Create acknowledgement commands like this:
        #
        #   echo TF2MON-KICKS-POP
        #   echo TF2MON-KICKS-POPLEFT
        #   echo TF2MON-SPAMS-POP
        #   echo TF2MON-SPAMS-POPLEFT

        if self.msgs:
            last, first = self.msgs[-1], self.msgs[0]
            last_ack = f" ; echo {APPTAG}{self.name.upper()}-POP"
            first_ack = last_ack + "LEFT"
        else:
            _echo = "say" if tf2mon.DebugFlagControl.value else "echo"
            last = first = f"{_echo} the {self.name} queue is empty."
            last_ack = first_ack = ""

        return [
            f'alias _tf2mon_{self.name}_pop "{last}{last_ack}"',
            f'alias _tf2mon_{self.name}_popleft "{first}{first_ack}"',
        ]



from typing import Match

from loguru import logger

import tf2mon
from tf2mon.gameevent import GameEvent
from tf2mon.user import UserKey


class GamePerkOnEvent(GameEvent):

    pattern = (
        r"[0-9A-F]{6}\[RTD\] [0-9A-F]{6}(?P<username>.*) rolled [0-9A-F]{6}(?P<perk>.*)"
    )

    def handler(self, match: Match[str]) -> None:

        username, perk = match.groups()
        user = tf2mon.users[UserKey(username)]
        user.perk = perk
        user.dirty = True
        logger.log("PERK-ON", f"{user} {perk!r}")


class GamePerkOff1Event(GameEvent):

    pattern = r"[0-9A-F]{6}\[RTD\] [0-9A-F]{6}(?P<username>.*)\'s perk has worn off."

    def handler(self, match: Match[str]) -> None:

        (username,) = match.groups()
        user = tf2mon.users[UserKey(username)]
        user.perk = ""
        user.dirty = True
        logger.log("PERK-OFF", f"{user} {user.perk!r}")


class GamePerkOff2Event(GameEvent):

    pattern = r"[0-9A-F]{6}\[RTD\] Your perk has worn off."

    def handler(self, _match: Match[str] | None) -> None:

        user = tf2mon.users.me
        user.perk = ""
        user.dirty = True
        logger.log("PERK-OFF", f"{user} {user.perk!r}")


class GamePerkChangeEvent(GameEvent):

    pattern = r"[0-9A-F]{6}\[RTD\] [0-9A-F]{6}(?P<username>.*) has changed class during their roll."

    def handler(self, match: Match[str]) -> None:

        (username,) = match.groups()
        user = tf2mon.users[UserKey(username)]
        logger.debug(f"{user} changed class")



"""MsgQueue control."""

from tf2mon.control import Control
from tf2mon.msgqueue import MsgQueue


class MsgQueueControl(Control):
    """MsgQueue control."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self) -> None:
        """Initialize control."""

        self.msgq = MsgQueue(self.name)
        self.msgs = self.msgq.msgs
        self.clear = self.msgq.clear
        self.pop = self.msgq.pop
        self.popleft = self.msgq.popleft
        self.push = self.msgq.push
        self.pushleft = self.msgq.pushleft
        self.aliases = self.msgq.aliases
        super().__init__()



"""Cycle scoreboard Sort column."""

from argparse import ArgumentParser
from enum import Enum
from typing import Match

import tf2mon
from tf2mon.control import CycleControl
from tf2mon.cycle import Cycle


class SortOrderControl(CycleControl):
    """Cycle scoreboard Sort column."""

    name = "TOGGLE-SORT"
    enum = Enum("enum", "AGE STEAMID CONN K KD USERNAME")
    cycle = Cycle("_t_soc", enum)
    items = {
        enum.AGE: lambda user: (user.age, user.username_upper),
        enum.STEAMID: lambda user: (user.steamid.id if user.steamid else 0, user.username_upper),
        enum.K: lambda user: (-user.nkills, user.username_upper),
        enum.KD: lambda user: (-user.kdratio, -user.nkills, user.username_upper),
        enum.CONN: lambda user: (user.elapsed, user.username_upper),
        enum.USERNAME: lambda user: user.username_upper,
    }

    def start(self) -> None:
        self.cycle.start(self.enum.__dict__[tf2mon.options.sort_order])
        tf2mon.ui.scoreboard.set_sort_order(self.cycle.value.name)
        assert self.cycle.value.name == tf2mon.options.sort_order

    def handler(self, _match: Match[str] | None) -> None:
        _ = self.cycle.next
        tf2mon.ui.scoreboard.set_sort_order(self.cycle.value.name)
        tf2mon.ui.update_display()

    def add_arguments_to(self, parser: ArgumentParser) -> None:
        arg = parser.add_argument(
            "--sort-order",
            choices=[x.name for x in list(self.enum)],
            default="KD",
            help="choose sort order",
        )
        self.add_fkey_to_help(arg)
        self.cli.add_default_to_help(arg)



"""Logger controls."""

from argparse import ArgumentParser
from enum import Enum
from typing import Match

from loguru import logger as _logger

import tf2mon
from tf2mon.control import Control, CycleControl
from tf2mon.cycle import Cycle


class LogLevelControl(CycleControl):
    """Cycle logger `level`."""

    name = "TOGGLE-LOG-LEVEL"
    enum = Enum("enum", "INFO DEBUG TRACE")
    cycle = Cycle("_t_loglvl", enum)
    items = {
        enum.INFO: "INFO",  # ""
        enum.DEBUG: "DEBUG",  # "-v"
        enum.TRACE: "TRACE",  # "-vv"
    }

    def start(self) -> None:
        """Set logging level based on `--verbose`."""

        tf2mon.ui.logsink.set_verbose(tf2mon.options.verbose)
        self.cycle.start(self.enum.__dict__[tf2mon.ui.logsink.level])

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.ui.logsink.set_level(self.items[self.cycle.next])
        tf2mon.ui.show_status()


class LogLocationControl(CycleControl):
    """Cycle logger `location` format."""

    name = "TOGGLE-LOG-LOCATION"
    enum = Enum("enum", "MOD NAM THM THN FILE NUL")
    cycle = Cycle("_t_logloc", enum)
    items = {
        enum.MOD: "{module}.{function}:{line}",
        enum.NAM: "{name}.{function}:{line}",
        enum.THM: "{thread.name}:{module}.{function}:{line}",
        enum.THN: "{thread.name}:{name}.{function}:{line}",
        enum.FILE: "{file}:{function}:{line}",
        enum.NUL: None,
    }

    def start(self) -> None:
        self.cycle.start(self.enum.__dict__[tf2mon.options.log_location])
        tf2mon.ui.logsink.set_location(self.items[self.cycle.value])

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.ui.logsink.set_location(self.items[self.cycle.cycle])
        tf2mon.ui.show_status()

    def add_arguments_to(self, parser: ArgumentParser) -> None:
        arg = parser.add_argument(
            "--log-location",
            choices=[x.name for x in list(self.enum)],
            default="NUL",
            help="choose format of logger location field",
        )
        self.add_fkey_to_help(arg)
        self.cli.add_default_to_help(arg)


class ResetPaddingControl(Control):
    """Reset logger `padding`."""

    name = "RESET-PADDING"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.ui.logsink.reset_padding()
        _logger.info("padding reset")



from typing import Match

from loguru import logger

import tf2mon
from tf2mon.chat import Chat
from tf2mon.gameevent import GameEvent
from tf2mon.player import Player
from tf2mon.racist import is_racist_text
from tf2mon.user import UserKey


class GameChatEvent(GameEvent):

    # 'Bob :  hello'
    # '*DEAD* Bob :  hello'
    # '*DEAD*(TEAM) Bob :  hello'

    pattern = (
        r"(?:(?P<dead>\*DEAD\*)?(?P<teamflag>\(TEAM\))? )?(?P<username>.*) :  ?(?P<msg>.*)$"
    )

    def handler(self, match: Match[str]) -> None:

        _dead, teamflag, username, msg = match.groups()

        user = tf2mon.users[UserKey(username)]
        chat = Chat(user, bool(teamflag), msg)

        user.chats.append(chat)
        tf2mon.ChatsControl.append(chat)

        level = "TEAMCHAT" if chat.teamflag else "CHAT"
        if user.team:
            level += user.team.name
        logger.log(level, f"{user} {msg}")

        # if this is a team chat, then we know we're on the same team, and
        # if one of us knows which team we're on and the other doesn't, we
        # can assign.

        if chat.teamflag and user != tf2mon.users.me:
            # we're on the same team
            if not user.team:
                if tf2mon.users.my.team:
                    user.team = tf2mon.users.my.team
            elif not tf2mon.users.my.team:
                tf2mon.users.my.team = user.team

        # inspect msg
        if is_racist_text(chat.msg):
            user.kick(Player.RACIST)

        elif user.is_cheater_chat(chat):
            user.kick(Player.CHEATER)



"""Spams queue control."""

from typing import Match

import tf2mon
from tf2mon.controls.msgqueue import MsgQueueControl


class SpamsControl(MsgQueueControl):
    """Spams queue control."""

    name = "spams"


class SpamsPopControl(SpamsControl):
    """Pop last/latest spams queue message."""

    name = "SPAMS-POP"
    action = "tf2mon_spams_pop"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.SpamsControl.pop()
        tf2mon.ui.refresh_spams()


class SpamsClearControl(SpamsControl):
    """Clear spams queue."""

    name = "SPAMS-CLEAR"
    action = "tf2mon_spams_clear"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.SpamsControl.clear()
        tf2mon.ui.refresh_spams()


class SpamsPopleftControl(SpamsControl):
    """Pop first/oldest spams queue message."""

    name = "SPAMS-POPLEFT"
    action = "tf2mon_spams_popleft"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.SpamsControl.popleft()
        tf2mon.ui.refresh_spams()



"""Docstring."""

import re
from typing import Match, Pattern


class GameEvent:
    """Base class of all game events."""

    pattern: str
    start_stepping = False  # pause before handling
    _re: Pattern[str]

    def __init__(self) -> None:
        """Docstring."""

        if self.pattern:
            self._re = re.compile(self.pattern)
            self.match = self._re.match
            self.search = self._re.search
        # else:
        #     self.match = lambda _line: False
        #     self.search = lambda _line: False

    # def match(self, _line: str) -> re.Match:
    #     return None

    # def search(self, _line: str) -> re.Match:
    #     return None

    def handler(self, match: Match[str]) -> None:
        """Docstring."""

        raise NotImplementedError



"""Interface to `ISteamUser.GetPlayerSummaries`."""

import time

import steam.webapi  # type: ignore
from loguru import logger

from tf2mon.steamid import BOT_STEAMID, SteamID
from tf2mon.steamplayer import SteamPlayer

MAX_AGE = 2 * 60 * 60


class SteamWebAPI:
    """Interface to `ISteamUser.GetPlayerSummaries`.

    Results are cached to avoid banging the server.
    """

    def __init__(self, webapi_key: str):
        """Initialize interface."""

        if webapi_key:
            self._webapi = steam.webapi.WebAPI(key=webapi_key)
        else:
            self._webapi = None
            logger.warning("Running without `webapi_key`")

        self._nbots = 0

    def fetch_steamid(self, steamid: int) -> SteamPlayer:
        """Lookup and return `SteamPlayer` with matching `steamid`.

        Use web service to get "Player Summary" of given `steamid`.
        Create dummy object for game bots.
        """

        now = int(time.time())

        if steamid == BOT_STEAMID.id:
            # create a dummy steamid for this bot; (not a hacker, a real game bot)
            self._nbots += 1
            return SteamPlayer(
                steamid=steamid,
                personaname="",
                profileurl="",
                personastate=0,
                realname="",
                timecreated=now - (self._nbots * 86400),
                loccountrycode="US",
                locstatecode="IL",
                loccityid="CHGO",
            )

        if not SteamID(steamid).is_valid():
            return SteamPlayer(steamid=steamid, personaname="???", timecreated=now, mtime=now)

        # it's not a game bot; look in database.
        steamplayer = SteamPlayer.fetch_steamid(steamid)

        if steamplayer and steamplayer.mtime > now - MAX_AGE:
            logger.debug("current")
            return steamplayer
        if steamplayer:
            logger.debug("expired")

        # not current or not in cache; call web service.
        player_summaries = self._get_player_summaries([SteamID(steamid)])

        if len(player_summaries) < 1:
            return SteamPlayer(steamid=steamid, personaname="???", timecreated=now, mtime=now)

        summary = player_summaries[0]

        steamplayer = SteamPlayer(
            steamid,
            summary.get("personaname", ""),
            summary.get("profileurl", ""),
            int(summary.get("personastate", 0)),
            summary.get("realname", ""),
            int(summary.get("timecreated", 0)),
            summary.get("loccountrycode", ""),
            summary.get("locstatecode", ""),
            summary.get("loccityid", ""),
            now,
        )

        steamplayer.upsert()
        return steamplayer

    def _get_player_summaries(self, steamids: list[SteamID]) -> list[dict[str, str]]:

        if not self._webapi:
            return []

        jdoc = self._webapi.call(
            "ISteamUser.GetPlayerSummaries", steamids=",".join([str(x.as_64) for x in steamids])
        )

        return list(jdoc["response"]["players"])



import pytest

from tf2mon.cycle import Cycle


def test_prev() -> None:
    cycle = Cycle("numbers", ["one", "two", "three", "four", "five"])
    assert cycle.value == "one"
    assert cycle.prev == "five"
    assert cycle.prev == "four"
    assert cycle.prev == "three"
    assert cycle.prev == "two"
    assert cycle.prev == "one"
    assert cycle.prev == "five"
    assert cycle.value == "five"


def test_next() -> None:
    cycle = Cycle("numbers", ["one", "two", "three", "four", "five"])
    assert cycle.value == "one"
    assert cycle.next == "two"
    assert cycle.next == "three"
    assert cycle.next == "four"
    assert cycle.next == "five"
    assert cycle.next == "one"
    assert cycle.next == "two"
    assert cycle.value == "two"


def test_getitem() -> None:
    cycle = Cycle("numbers", ["one", "two", "three", "four", "five"])
    assert cycle[0] == "one"
    assert cycle[1] == "two"
    assert cycle[2] == "three"
    assert cycle[3] == "four"
    assert cycle[4] == "five"
    assert cycle[-1] == "five"
    assert cycle[-5] == "one"

    with pytest.raises(IndexError):
        _ = cycle[5]
    with pytest.raises(IndexError):
        _ = cycle[-6]


def test_start() -> None:
    cycle = Cycle("numbers", ["one", "two", "three", "four", "five"])
    cycle.start("one")
    assert cycle.value == "one"
    cycle.start("two")
    assert cycle.value == "two"
    cycle.start("three")
    assert cycle.prev == "two"
    cycle.start("four")
    assert cycle.next == "five"
    cycle.start("five")
    assert cycle.value == "five"

    with pytest.raises(KeyError):
        cycle.start("zero")
    with pytest.raises(KeyError):
        cycle.start("")
    with pytest.raises(KeyError):
        cycle.start(None)



"""Kick last killer."""

from typing import Match

import tf2mon
from tf2mon.control import Control
from tf2mon.player import Player


class KickLastControl(Control):
    """Kick last killer."""

    attr: str

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.users.kick_my_last_killer(self.attr)

    def status(self) -> str:
        return self.attr


class KickLastCheaterControl(KickLastControl):
    """Kick last killer as `cheater`."""

    name = "KICK-LAST-CHEATER"
    attr = Player.CHEATER


class KickLastRacistControl(KickLastControl):
    """Kick last killer as `racist`."""

    name = "KICK-LAST-RACIST"
    attr = Player.RACIST


class KickLastSuspectControl(KickLastControl):
    """Mark last killer as `suspect`."""

    name = "KICK-LAST-SUSPECT"
    attr = Player.SUSPECT
